﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Linq;
using System.Data;
using System.Collections.ObjectModel;
using OptimisationManagerExtention.ViewExtention;
using OptimisationManagerExtention.Model.FileReaders;

using OptimisationManagerExtention.Model.Terminal;
using System.IO;
using OptimisationManagerExtention.Model.OptimisationManager;
using OptimisationManagerExtention.Model.OptimisationManager.Managers;
using OptimisationManagerExtention.Model.DirectoryManagers;

namespace OptimisationManagerExtention.Model
{
    /// <summary>
    /// Фабрика по подстановки модели в графический интерфейс
    /// </summary>
    class ModelCreator
    {
        /// <summary>
        /// Модель
        /// </summary>
        private static IExtentionGUI_M testModel;

        /// <summary>
        /// Property возвращающее либо модель (если она не была подменена) либо побмененную (для тестов) модель
        /// </summary>
        internal static IExtentionGUI_M Model => testModel ?? new ExtentionGUI_M(new MainTerminalCreator(),
                                                                                 new MainConfigCreator(),
                                                                                 new MainReportReaderCreator(),
                                                                                 new MainSetFileManagerCreator(),
                                                                                 new OptimisationExtentionWorkingDirectory("OptimisationManagerExtention"),
                                                                                 new MainOptimisatorSettingsManagerCreator(),
                                                                                 new TerminalDirectory());

        /// <summary>
        /// Метод подмены модели. подставляет тестовую модель что бы можно было отдельно тестировать графику от логики
        /// </summary>
        /// <param name="model">тестовая модель - подставляется извне</param>
        [System.Diagnostics.Conditional("DEBUG")]
        public static void SetModel(IExtentionGUI_M model)
        {
            testModel = model;
        }
    }
    class ExtentionGUI_M : IExtentionGUI_M
    {
        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="TerminalCreator">Фабрика по созданию менеджеров терминалла</param>
        /// <param name="ConfigCreator">Фабрика по созданию менеджеров конфигурационных файлов</param>
        /// <param name="ReportReaderCreator">Фабрика по созданию читателей отчетов</param>
        /// <param name="SetFileManagerCreator">Фабрика по созданию файлов с настройками терминаллов</param>
        public ExtentionGUI_M(TerminalCreator TerminalCreator,
                              ConfigCreator ConfigCreator,
                              ReportReaderCreator ReportReaderCreator,
                              SetFileManagerCreator SetFileManagerCreator,
                              OptimisationExtentionWorkingDirectory CurrentWorkingDirectory,
                              OptimisatorSettingsManagerCreator SettingsManagerCreator,
                              TerminalDirectory terminalDirectory)
        {
            // Присваиваем текущую рабочию директорию
            this.CurrentWorkingDirectory = CurrentWorkingDirectory;
            this.terminalDirectory = terminalDirectory;
            //Создаем фабрики
            this.TerminalCreator = TerminalCreator;
            this.ReportReaderCreator = ReportReaderCreator;
            this.ConfigCreator = ConfigCreator;
            this.SetFileManagerCreator = SetFileManagerCreator;
            this.SettingsManagerCreator = SettingsManagerCreator;
            CreateOptimisationManagerFabrics();

            // подписываемся на событие обновления коллекции колонок в таблице с историческими оптимизациями
            HistoryOptimisationResults.Columns.CollectionChanged += Columns_CollectionChanged;

            // Присваиваем начальный статус
            Status = new VarKeeper<string>("Status", "Wait for the operation");
            Status.PropertyChanged += OnPropertyChanged;
            // Присваиваем начальные значения прогресс бара
            PB_Value = new VarKeeper<double>("PB_Value", 0);
            PB_Value.PropertyChanged += OnPropertyChanged;
            // Создаем переменную зранящую в себе индекс выбранного терминалла из списка доступных терминаллов на которых была оптимизация
            TerminalsAfterOptimisation_Selected = new VarKeeper<int>("TerminalsAfterOptimisation_Selected", 0);
            TerminalsAfterOptimisation_Selected.PropertyChanged += OnPropertyChanged;
            // Создаем переменную зранящую в себе индекс выбранного робота из списка доступных роботов на которых была оптимизация
            BotsAfterOptimisation_Selected = new VarKeeper<int>("BotsAfterOptimisation_Selected", -1);
            BotsAfterOptimisation_Selected.PropertyChanged += OnPropertyChanged;

            _isTerminalsEnabled = new VarKeeper<bool>("IsTerminalsLVEnabled", true);
            _isTerminalsEnabled.PropertyChanged += OnPropertyChanged;

            // Загружаем сведения о терминаллах установленных на компьютере
            FillInTerminalsID();
            FillInTerminalsAfterOptimisation();
            LoadOptimisations();
        }

        /// <summary>
        /// Переменная содержащая в себе сведения о директории исспользуемой данным дополнением 
        /// для хранения / изменения собственных вспомогательных файлов
        /// </summary>
        protected readonly OptimisationExtentionWorkingDirectory CurrentWorkingDirectory;
        protected readonly TerminalDirectory terminalDirectory;
        /// <summary>
        /// Деструктор
        /// </summary>
        ~ExtentionGUI_M()
        {
            // Отписываемся от событий на которые подписывались в конструкторе
            HistoryOptimisationResults.Columns.CollectionChanged -= Columns_CollectionChanged;
            Status.PropertyChanged -= OnPropertyChanged;
            PB_Value.PropertyChanged -= OnPropertyChanged;
            TerminalsAfterOptimisation_Selected.PropertyChanged -= OnPropertyChanged;
            BotsAfterOptimisation_Selected.PropertyChanged -= OnPropertyChanged;
            _isTerminalsEnabled.PropertyChanged -= OnPropertyChanged;
            // Отписываемся от события на которое подписались во время начала оптимизации (если таковая подписка существует)
            if (selectedOptimisationManager != null)
                selectedOptimisationManager.AllOptimisationsFinished -= SelectedOptimisationManager_AllOptimisationsFinished;
        }

        #region Fabrics
        /// <summary>
        /// Фабрика создающая менеджеры терминаллов
        /// </summary>
        protected readonly TerminalCreator TerminalCreator;
        /// <summary>
        /// Фабрика создающая менеджеры файлов конфигураций
        /// </summary>
        protected readonly ConfigCreator ConfigCreator;
        /// <summary>
        /// Фабрика создающая менеджеры чтения отчетов
        /// </summary>
        protected readonly ReportReaderCreator ReportReaderCreator;
        /// <summary>
        /// Фабрика создающая менеджеры для работы с файлами настройки терминаллов
        /// </summary>
        protected readonly SetFileManagerCreator SetFileManagerCreator;
        /// <summary>
        /// Файбрика создающая менеджеры для управления процессами оптимизаций
        /// </summary>
        protected List<OptimisationManagerFabric> optimisationManagerFabrics;
        /// <summary>
        /// Фабрика создания менеджера по работе с файлами оптимизации
        /// </summary>
        protected readonly OptimisatorSettingsManagerCreator SettingsManagerCreator;
        #endregion

        /// <summary>
        /// конкретный выбраный на данный момент менеджер оптимизации
        /// </summary>
        private OptimisationManager.OptimisationManager selectedOptimisationManager = null;

        #region BotParams
        /// <summary>
        /// Представление таблицы выбранных терминаллов со стартовой вкладки данного дополнения
        /// key - Termonal ID
        /// value - bot params
        /// </summary>
        private readonly Dictionary<string, BotParamKeeper> BotParamsKeeper = new Dictionary<string, BotParamKeeper>();
        /// <summary>
        /// Выбранный терминалл на текущий момент
        /// </summary>
        private string selectedTerminalID = null;
        /// <summary>
        /// Список Параметров робота для ребактирования
        /// </summary>
        List<ParamsItem> IExtentionGUI_M.BotParams
        {
            get
            {
                return (BotParamsKeeper.Count > 0 && selectedTerminalID != null) ?
                       BotParamsKeeper[selectedTerminalID].BotParams.Params :
                       new List<ParamsItem>();
            }
        }
        /// <summary>
        /// Создание и получение ссылки на файл с настройками для оптимизации конкретного выбранного робота
        /// </summary>
        /// <param name="fullExpertName">Полное имя эксперта</param>
        /// <param name="TerminalID">ID терминалла</param>
        /// <returns>Конфигурированный менеджер работы с файлами настроек</returns>
        private SetFileManager GetSetFile(string fullExpertName, string TerminalID)
        {
            DirectoryInfo terminalChangableFolder = terminalDirectory.Terminals.Find(x => x.Name == TerminalID);

            // Создаем менеджер по работе с терминалламом
            ITerminalManager terminalManager = TerminalCreator.Create(terminalChangableFolder);

            // Составляем путь до папки Tester (которая находится в директории ~/MQL5/Profiles) 
            // Если ее нет (иногда метатрейдер не создает ее при первой установке) то создаем ее самостоятельно
            // В ней хранятся файлы с настройками оптимизационных параметров
            DirectoryInfo pathToMqlTesterFolder = terminalManager.MQL5Directory.GetDirectory("Profiles").GetDirectory("Tester", true);
            if (pathToMqlTesterFolder == null)
                throw new Exception("Can`t find (or create) ~/MQL5/Profiles/Tester directory");

            // Создаем конфигурационный файл и сразу копируем его в папку Configs текущей рабочей директори данного дополнения
            Config config = ConfigCreator.Create(Path.Combine(terminalChangableFolder.GetDirectory("config").FullName, "common.ini"))
                                         .DublicateFile(Path.Combine(CurrentWorkingDirectory.Configs.FullName, $"{TerminalID}.ini"));
            // Настраиваем терминалл так чтобы он запустил тест выбранного робота и сразу же выключился
            // Это необходимо для того что бы терминалл создалл .set файл с настройками данного советника.
            // Для того что бы он сразу же выключился - указываем дату окончания теста на день ниже чем дата начала.
            config.Tester.Expert = fullExpertName;
            config.Tester.Model = ENUM_Model.OHLC_1_minute;
            config.Tester.Optimization = ENUM_OptimisationMode.Disabled;
            config.Tester.Period = ENUM_Timeframes.D1;
            config.Tester.ShutdownTerminal = true;
            config.Tester.FromDate = DateTime.Now.Date;
            config.Tester.ToDate = config.Tester.FromDate.Value.AddDays(-1);

            // Задаем менеджеру терминалла конфигурационный файл, запускаем его и дожижаемся закрытия терминалла
            // Для того что ды терминалл закрылся автоматически после завершения теста, 
            // приспаеваем полю config.Tester.ShutdownTerminal значение  true
            terminalManager.Config = config;
            terminalManager.WindowStyle = System.Diagnostics.ProcessWindowStyle.Minimized;
            string fileName = $"{new FileInfo(fullExpertName).Name.Split('.')[0]}.set";

            while (!terminalManager.Run())
            {
                System.Windows.MessageBoxResult mb_ans =
                    System.Windows.MessageBox.Show(@"Can`t start terminal 
Close by hands all Metatrader terminals that running now (except main terminal)",
"Can`t start terminal", System.Windows.MessageBoxButton.OKCancel);
                if (mb_ans == System.Windows.MessageBoxResult.Cancel)
                    break;
            }
            terminalManager.WaitForStop();

            bool isSetFileWasCreated = pathToMqlTesterFolder.GetFiles().Any(x => x.Name == fileName);

            return SetFileManagerCreator.Create(Path.Combine(pathToMqlTesterFolder.FullName, fileName), !isSetFileWasCreated);
        }
        /// <summary>
        /// Добавляем новый терминалл в список терминаллов для оптимизации 
        /// </summary>
        /// <param name="fullExpertName">Полное имя робота относительно папки Experts</param>
        /// <param name="TerminalID">ID выбранного терминалла</param>
        private void AddNewTerminalIntoBotParamsKeeper(string fullExpertName, string TerminalID)
        {
            // Собираем структуру хранящую в себе параметры робота
            BotParamKeeper botParam = new BotParamKeeper
            {
                BotName = fullExpertName,
                BotParams = GetSetFile(fullExpertName, TerminalID),
                OptimisationData = new OptimisationInputData()
            };
            botParam.OptimisationData.ForvardDate = DateTime.Now.Date;

            // Добавляем новый эллемент в коллекцию
            BotParamsKeeper.Add(TerminalID, botParam);
        }
        /// <summary>
        /// Данный метод вызывается при выборе эксперта для оптимизации / теста 
        /// из таблицы с терминаллами на вкладке Settings
        /// 
        /// Во время его работы :
        /// 1. Сохранятся в прошлом выбранном роботе его оптимизационные настройки - "optimisationInputData"
        /// 2. Загружаются параметры нововыбранного робота в таблицу "BotParams"
        /// 3. в переменную "optimisationInputData" - передается значения настроек оптимизации 
        ///    для нововыбранного робота
        /// </summary>
        /// <param name="fullExpertName">Путь к эксперту относительно папки ~/Experts</param>
        /// <param name="TerminalID">ID запускаемого терминалла</param>
        /// <param name="optimisationInputData">настройки оптимизации</param>
        void IExtentionGUI_M.LoadBotParams(string fullExpertName,
            string TerminalID,
            out OptimisationInputData? optimisationInputData)
        {
            PBUpdate(0, "Loading params", true);
            optimisationInputData = null;

            if (!IsTerminalsLVEnabled)
                return;

            _isTerminalsEnabled.SetVar(false);

            if (!BotParamsKeeper.Keys.Contains(TerminalID))
            {
                PBUpdate(100, "Add New Terminal", false);
                AddNewTerminalIntoBotParamsKeeper(fullExpertName, TerminalID);
            }
            else
            {
                if (selectedTerminalID != null)
                    BotParamsKeeper[selectedTerminalID].BotParams.SaveParams();
                else
                {
                    foreach (var item in BotParamsKeeper)
                    {
                        item.Value.BotParams.SaveParams();
                    }
                }
            }

            selectedTerminalID = TerminalID;
            optimisationInputData = BotParamsKeeper[selectedTerminalID].OptimisationData;

            if (BotParamsKeeper[selectedTerminalID].BotName != fullExpertName)
            {
                PBUpdate(100, "Load new params", false);
                BotParamKeeper param = BotParamsKeeper[selectedTerminalID];
                param.BotName = fullExpertName;
                param.BotParams = GetSetFile(fullExpertName, TerminalID);
                BotParamsKeeper[selectedTerminalID] = param;
            }
            PBUpdate(0, null, true);
            _isTerminalsEnabled.SetVar(true);
        }
        /// <summary>
        /// Вызывается при удалении переданного терминалла из списка терминаллов
        /// В методе удаляются параметры для выбранного робота в данном терминале
        /// </summary>
        /// <param name="item">выбранный для удаления терминалл</param>
        bool IExtentionGUI_M.RemoveBotParams(string TerminalID)
        {
            if (!IsTerminalsLVEnabled)
                return false;

            if (BotParamsKeeper.Keys.Contains(TerminalID))
            {
                BotParamsKeeper.Remove(TerminalID);
                if (selectedTerminalID == TerminalID)
                    selectedTerminalID = null;
            }
            return true;
        }
        /// <summary>
        /// Проверка на то возможно ли сейчас добавлять новый терминалл в список терминаллов отведенных для оптимизации
        /// </summary>
        /// <returns>true если возможно</returns>
        bool IExtentionGUI_M.IsEnableToAddNewTerminal()
        {
            return IsTerminalsLVEnabled;
        }
        void IExtentionGUI_M.UpdateTerminalOptimisationsParams(OptimisationInputData optimisationInputData)
        {
            if (selectedTerminalID != null)
            {
                BotParamKeeper paramKeeper = BotParamsKeeper[selectedTerminalID];
                paramKeeper.OptimisationData.Copy(optimisationInputData);
                BotParamsKeeper[selectedTerminalID] = paramKeeper;
            }
        }
        #endregion

        #region ProgressBar and Status
        /// <summary>
        /// Статус выполняемой задачи
        /// </summary>
        public VarKeeper<string> Status { get; }
        /// <summary>
        /// Значение програсс бара
        /// </summary>
        public VarKeeper<double> PB_Value { get; }
        /// <summary>
        /// Метод обновляющий значение прогресс бара и статус выполняемой операции
        /// </summary>
        /// <param name="increasing_item">число на которое увеличится прогресс бар</param>
        /// <param name="status">Статус выполняемой операции</param>
        /// <param name="isClear">Признак стоит ли отчистить прогресс бар</param>
        protected virtual void PBUpdate(double increasing_item, string status, bool isClear)
        {
            MQL5Connector.CurrentDispatcher.Invoke(() =>
            {
                string _status = (string.IsNullOrEmpty(status) || string.IsNullOrWhiteSpace(status)) ? "Wait for the operation" : status;
                if (isClear)
                    PB_Value.SetVar(0);
                else
                {
                    double val = PB_Value + increasing_item;
                    PB_Value.SetVar(val);
                }
                Status.SetVar(_status);
            });
        }
        #endregion

        /// <summary>
        /// ID терминалов на стартавой странице
        /// определяются при старте программы и неизменны до ее завершения 
        /// </summary>
        public ObservableCollection<string> TerminalsID { get; } = new ObservableCollection<string>();

        #region Optimisation results data
        /// <summary>
        /// Таблица с результатами исторической оптимизации
        /// </summary>
        public DataTable HistoryOptimisationResults { get; } = new DataTable();
        /// <summary>
        /// Таблица с результатами форвардной оптимизации
        /// </summary>
        public DataTable ForvardOptimisationResults { get; } = new DataTable();
        /// <summary>
        /// Заголовки в таблицах с результатами оптимизаций
        /// </summary>
        public ObservableCollection<ColumnDescriptor> OptimisationResultsColumnHeadders { get; } = new ObservableCollection<ColumnDescriptor>();
        /// <summary>
        /// Список терминаллов на которых проходила оптимизация
        /// </summary>
        public ObservableCollection<string> TerminalsAfterOptimisation { get; } = new ObservableCollection<string>();
        /// <summary>
        /// Индекс выбранного терминалла на котором прозодила оптимизация
        /// </summary>
        public VarKeeper<int> TerminalsAfterOptimisation_Selected { get; set; }
        /// <summary>
        /// Уникальный список роботов прошедших оптимизацию
        /// </summary>
        public ObservableCollection<string> BotsAfterOptimisation { get; } = new ObservableCollection<string>();
        /// <summary>
        /// Индекс выбранного робота который прошел оптимизацию
        /// </summary>
        public VarKeeper<int> BotsAfterOptimisation_Selected { get; set; }
        /// <summary>
        /// Событие срабатывающее после изменения списка колонок в таблице с историческими оптимизациями
        /// </summary>
        /// <param name="sender"></param>
        /// <param name="e"></param>
        private void Columns_CollectionChanged(object sender, CollectionChangeEventArgs e)
        {
            switch (e.Action)
            {
                // Добавление колонок в список колонок для таблиц
                case CollectionChangeAction.Add:
                    {
                        foreach (DataColumn item in HistoryOptimisationResults.Columns)
                        {
                            if (!OptimisationResultsColumnHeadders.Any(x => x.DisplayMember == item.ColumnName))
                            {
                                OptimisationResultsColumnHeadders.Add(new ColumnDescriptor
                                {
                                    HeaderText = item.ColumnName,
                                    DisplayMember = item.ColumnName
                                });
                            }
                        }
                    }
                    break;
                // Удаление колонок из списка колонок для таблиц
                case CollectionChangeAction.Remove:
                    {
                        List<string> ColNames = new List<string>(OptimisationResultsColumnHeadders.Select(x => x.DisplayMember));
                        var TableColNames = from DataColumn c in HistoryOptimisationResults.Columns
                                            select c.ColumnName;

                        foreach (var item in ColNames)
                        {
                            if (!TableColNames.Contains(item))
                            {
                                ColumnDescriptor descriptor = OptimisationResultsColumnHeadders.
                                     First(x => x.DisplayMember == item);
                                OptimisationResultsColumnHeadders.Remove(descriptor);
                            }
                        }
                    }
                    break;
                // Обновление колонок
                default:
                    {
                        OptimisationResultsColumnHeadders.Clear();
                        var TableColNames = from DataColumn c in HistoryOptimisationResults.Columns
                                            select c.ColumnName;

                        foreach (var item in TableColNames)
                        {
                            OptimisationResultsColumnHeadders.Add(new ColumnDescriptor
                            {
                                HeaderText = item,
                                DisplayMember = item
                            });
                        }
                    }
                    break;
                    //default:
                    //    break;
            }
        }
        /// <summary>
        /// Заполнение списка роботов для для нововыбранного терминалла
        /// </summary>
        public void SelectNewBotsAfterOptimisation_forNewTerminal()
        {
            BotsAfterOptimisation.Clear();
            BotsAfterOptimisation_Selected.SetVar(0);

            HistoryOptimisationResults.Clear();
            HistoryOptimisationResults.Columns.Clear();
            ForvardOptimisationResults.Clear();
            ForvardOptimisationResults.Columns.Clear();

            if (TerminalsAfterOptimisation.Count == 0)
            {
                return;
            }

            string TerminalID = TerminalsAfterOptimisation[TerminalsAfterOptimisation_Selected];

            DirectoryInfo terminalChangableFolder = terminalDirectory.Terminals.Find(x => x.Name == TerminalID);
            ITerminalManager terminal = TerminalCreator.Create(terminalChangableFolder);

            IEnumerable<DirectoryInfo> directories = CurrentWorkingDirectory.Reports.GetDirectory(TerminalID).GetDirectories();

            foreach (var item in directories)
            {
                using (OptimisatorSettingsManager settingsManager = SettingsManagerCreator.Create(item.FullName))
                {
                    if (!terminal.Experts.Contains(settingsManager.GetItem(NodeType.ConfigCreator_inputData, "pathToBot")))
                    {
                        item.Delete(true);
                        continue;
                    }
                    BotsAfterOptimisation.Add(item.Name);
                }
            }
        }

        /// <summary>
        /// Коллбек события завершения оптимизации в нем заполняем списки терминаллов и роботов прошедших оптимизацию
        /// </summary>
        /// <param name="optimisationManager"></param>
        private void SelectedOptimisationManager_AllOptimisationsFinished(OptimisationManager.OptimisationManager optimisationManager)
        {
            MQL5Connector.CurrentDispatcher.Invoke(() =>
            {
                FillInTerminalsAfterOptimisation();
                LoadOptimisations();
            });
        }

        /// <summary>
        /// Метод заполняющий таблицы с оптимизационными результатами
        /// </summary>
        public void LoadOptimisations()
        {
            // Внутренний метод заполняющий переданную в него таблицу данными
            void SetData(bool isForvard, DataTable tb)
            {
                // Отчищаем таблицу от ранее заполненных данных
                tb.Clear();
                tb.Columns.Clear();

                // Получаем данные
                string TerminalID = TerminalsAfterOptimisation[TerminalsAfterOptimisation_Selected];
                string botName = BotsAfterOptimisation[BotsAfterOptimisation_Selected];
                string path = Path.Combine(CurrentWorkingDirectory.Reports
                                                                  .GetDirectory(TerminalID)
                                                                  .GetDirectory(botName)
                                                                  .FullName,
                                                                  $"{(isForvard ? "Forward" : "History")}.xml");
                if (!File.Exists(path))
                    return;

                using (ReportReader reader = ReportReaderCreator.Create(path))
                {
                    if (reader.ColumnNames.Count == 0)
                        return;

                    // Заполняем колонки
                    foreach (var item in reader.ColumnNames)
                    {
                        tb.Columns.Add(item);
                    }

                    // Заполняем строки
                    while (reader.Read(out List<KeyValuePair<string, object>> data))
                    {
                        DataRow row = tb.NewRow();
                        foreach (var item in data)
                        {
                            row[item.Key] = item.Value;
                        }
                        tb.Rows.Add(row);
                    }
                }
            }

            if (TerminalsAfterOptimisation.Count == 0 && BotsAfterOptimisation.Count == 0)
            {
                return;
            }

            // Заполняем поочередно сначала исторические а затем и форвардные оптимизации
            SetData(false, HistoryOptimisationResults);
            SetData(true, ForvardOptimisationResults);
        }
        /// <summary>
        /// Признак того доступна ли таблица с терминаллами или же нет
        /// </summary>
        private readonly VarKeeper<bool> _isTerminalsEnabled;
        public bool IsTerminalsLVEnabled
        {
            get
            {
                return _isTerminalsEnabled && (selectedOptimisationManager == null || !selectedOptimisationManager.IsOptimisationOrTestInProcess);
            }
        }
        #endregion

        #region PropertyChanged
        /// <summary>
        /// Событие изменения какого либо из свойств
        /// </summary>
        public event PropertyChangedEventHandler PropertyChanged;
        /// <summary>
        /// Метод запускающий событие изменения какого либо из свойств
        /// </summary>
        /// <param name="propertyName">Имя измененного свойства</param>
        protected virtual void OnPropertyChanged(string propertyName)
        {
            PropertyChanged?.Invoke(this, new PropertyChangedEventArgs(propertyName));
        }
        #endregion

        #region Initialisors
        /// <summary>
        /// Заполняем Терминаллы по которым существующи отчеты
        /// </summary>
        protected virtual void FillInTerminalsAfterOptimisation()
        {
            TerminalsAfterOptimisation.Clear();
            IEnumerable<DirectoryInfo> reportDirectories = CurrentWorkingDirectory.Reports.GetDirectories();
            List<DirectoryInfo> terminals = terminalDirectory.Terminals;
            foreach (var item in reportDirectories)
            {
                if (terminals.Any(x => x.Name == item.Name))
                    TerminalsAfterOptimisation.Add(item.Name);
                else
                    item.Delete(true);
            }
            TerminalsAfterOptimisation_Selected.SetVar(0);
            SelectNewBotsAfterOptimisation_forNewTerminal();
        }
        /// <summary>
        /// Создаем список фабрик для запуска различных комбинаций оптимизаций
        /// </summary>
        protected virtual void CreateOptimisationManagerFabrics()
        {
            optimisationManagerFabrics = new List<OptimisationManagerFabric>
            {
                new  SimpleOptimisationManager_Creator(terminalDirectory,
                                                       TerminalCreator,
                                                       ConfigCreator,
                                                       ReportReaderCreator,
                                                       SetFileManagerCreator,
                                                       CurrentWorkingDirectory,
                                                       PBUpdate,
                                                       SettingsManagerCreator)
            };
        }
        /// <summary>
        /// Получаем идентификационные номера терминаллов для Combobox с выбором конкретного терминалла
        /// </summary>
        protected virtual void FillInTerminalsID()
        {
            IEnumerable<string> terminals_id_collection = terminalDirectory.Terminals.Select(x => x.Name);

            foreach (var item in terminals_id_collection)
            {
                if (item == MQL5Connector.MainTerminalID)
                    continue;

                TerminalsID.Add(item);
            }
        }
        #endregion

        /// <summary>
        /// Метод получает полные имена роботов для выбранного терминалла относительно папки ~/Experts
        /// </summary>
        /// <param name="terminalIndex">Индекс выбранного терминалла</param>
        /// <returns>
        /// Список полных имен роботов(с учетом пути относительно папки ~/Experts)
        /// </returns>
        List<string> IExtentionGUI_M.GetBotNamesList(int terminalIndex)
        {
            ITerminalManager terminalManager = TerminalCreator.Create(terminalDirectory.Terminals.Find(x => x.Name == TerminalsID[terminalIndex]));
            return terminalManager.Experts;
        }

        /// <summary>
        /// Метод получает логин из мнмциализационного файла для выбранного терминалла
        /// </summary>
        /// <param name="terminalIndex">индекс выбранного терминалла</param>
        /// <returns>Логин терминалла</returns>
        uint? IExtentionGUI_M.GetCurrentLogin(int terminalIndex)
        {
            DirectoryInfo terminalChangableFolder = terminalDirectory.Terminals.Find(x => x.Name == TerminalsID[terminalIndex]);

            Config config = ConfigCreator.Create(Path.Combine(terminalChangableFolder.GetDirectory("config").FullName, "common.ini"));
            return config.Common.Login;
        }

        #region Start optimisation or test
        /// <summary>
        /// Метод запукскающий оптимизацию / тест по группе переданных советников
        /// </summary>
        /// <param name="SelectedTerminals"></param>
        async void IExtentionGUI_M.StartOptimisationOrTest(List<ViewModel.TerminalAndBotItem> SelectedTerminals)
        {
            if (BotParamsKeeper.Count == 0)
                return;
            foreach (var item in BotParamsKeeper)
            {
                item.Value.BotParams.SaveParams();
            }

            SetOptimisationManager(SelectedTerminals);

            // Запускаем оптимизацию и дожидаемся ее окончания
            _isTerminalsEnabled.SetVar(false);
            await System.Threading.Tasks.Task.Run(() => selectedOptimisationManager.StartOptimisation());
            _isTerminalsEnabled.SetVar(true);
        }

        private void SetOptimisationManager(List<ViewModel.TerminalAndBotItem> SelectedTerminals)
        {
            // Выбираем фабрику для создания оптимизационного менеджера из списка
            OptimisationManagerFabric OMFabric = optimisationManagerFabrics[0];
            //Отписываемся от ранее исспользованного оптимизационного менеджера
            if (selectedOptimisationManager != null)
            {
                // Проверяем запущена ли оптимизация в данный момент
                if (selectedOptimisationManager.IsOptimisationOrTestInProcess)
                    return;

                selectedOptimisationManager.AllOptimisationsFinished -= SelectedOptimisationManager_AllOptimisationsFinished;
            }

            // Создаем менеджера оптимизаций и подписываем его на событие заверщшения оптимизации
            selectedOptimisationManager = OMFabric.Create(BotParamsKeeper, SelectedTerminals);
            selectedOptimisationManager.AllOptimisationsFinished += SelectedOptimisationManager_AllOptimisationsFinished;
        }

        /// <summary>
        /// Запуск тестового прозода по событию двойного клика
        /// </summary>
        /// <param name="TableType">Тип таблицы с результатами оптимизаций</param>
        /// <param name="rowIndex">Индекс строки</param>
        async void IExtentionGUI_M.StartTest(ENUM_TableType TableType, int rowIndex)
        {
            if (!IsTerminalsLVEnabled)
                return;

            string TerminalID = TerminalsAfterOptimisation[TerminalsAfterOptimisation_Selected];
            string pathToBot = BotsAfterOptimisation[BotsAfterOptimisation_Selected];
            DirectoryInfo terminalChangableFolder = terminalDirectory.Terminals.Find(x => x.Name == TerminalID);

            DataRow row = (TableType == ENUM_TableType.History ? HistoryOptimisationResults : ForvardOptimisationResults).Rows[rowIndex];

            ConfigCreator_inputData configInputData;
            OptimisationInputData OptimisatorSettings;

            DirectoryInfo BotReportDirectory = CurrentWorkingDirectory.Reports.GetDirectory(TerminalID).GetDirectory(pathToBot);
            using (OptimisatorSettingsManager settingsManager = SettingsManagerCreator.Create(BotReportDirectory.FullName))
            {
                configInputData = settingsManager.ConfigCreator_inputData;
                OptimisatorSettings = settingsManager.OptimisationInputData;

                string setFilePath = Path.Combine(terminalChangableFolder
                                                  .GetDirectory("MQL5")
                                                  .GetDirectory("Profiles")
                                                  .GetDirectory("Tester", true).FullName,
                                                  configInputData.setFileName);

                SetFileManager setFile = SetFileManagerCreator.Create(setFilePath, true);
                setFile.Params = settingsManager.SetFileParams;

                foreach (var item in setFile.Params)
                {
                    if (row.Table.Columns.Contains(item.Variable))
                        item.Value = row[item.Variable].ToString();
                }
                setFile.SaveParams();
            }

            _isTerminalsEnabled.SetVar(false);
            if (selectedOptimisationManager == null)
                SetOptimisationManager(new List<ViewModel.TerminalAndBotItem>());

            await System.Threading.Tasks.Task.Run(() =>
            {
                selectedOptimisationManager.StartTest(configInputData, OptimisatorSettings);
            });
            _isTerminalsEnabled.SetVar(true);
        }

        #endregion
    }

    #region Accessory objects

    /// <summary>
    /// Структыра оптимизационных настроек робота
    /// </summary>
    struct OptimisationInputData
    {
        public void Copy(OptimisationInputData data)
        {
            Login = data.Login;
            ForvardDate = data.ForvardDate;
            IsVisual = data.IsVisual;
            Deposit = data.Deposit;
            Laverage = data.Laverage;
            Currency = data.Currency;
            DepositIndex = data.DepositIndex;
            ExecutionDelayIndex = data.ExecutionDelayIndex;
            ModelIndex = data.ModelIndex;
            CurrencyIndex = data.CurrencyIndex;
            LaverageIndex = data.LaverageIndex;
            OptimisationCriteriaIndex = data.OptimisationCriteriaIndex;
        }
        /// <summary>
        /// Логин видимый роботом в тестере / оптимизаторе
        /// </summary>
        public uint? Login;
        /// <summary>
        /// Дата начала форвардного периода
        /// </summary>
        public DateTime ForvardDate;
        /// <summary>
        /// Признак старта оптимизации в визуальном режиме
        /// </summary>
        public bool IsVisual;
        /// <summary>
        /// Депозит с которым стартует оптимизация
        /// </summary>
        public int Deposit;
        /// <summary>
        /// Кредитное плече
        /// </summary>
        public string Laverage;
        /// <summary>
        /// Валюта в которой измеряется прибыль
        /// </summary>
        public string Currency;
        /// <summary>
        /// Набор индексов для параметров оптимизации
        /// </summary>
        public int DepositIndex, ExecutionDelayIndex, ModelIndex,
                   CurrencyIndex, LaverageIndex, OptimisationCriteriaIndex;
        /// <summary>
        /// Виды подачи графика в робота при тестировании / оптимизации
        /// </summary>
        public ENUM_Model Model => GetEnum<ENUM_Model>(ModelIndex);
        /// <summary>
        /// Критерии оптимизационного отбора при оптимизации
        /// </summary>
        public ENUM_OptimisationCriteria GetOptimisationCriteria => GetEnum<ENUM_OptimisationCriteria>(OptimisationCriteriaIndex);
        /// <summary>
        /// Виды эмуляции задержек при исполнении ордера
        /// </summary>
        public ENUM_ExecutionDelay ExecutionDelay => GetEnum<ENUM_ExecutionDelay>(ExecutionDelayIndex);
        /// <summary>
        /// Преобразование индекса из массива имен перечисления в соответствующее перечисление
        /// </summary>
        /// <typeparam name="T">Тип перечисления к которому стоит привести индекс</typeparam>
        /// <param name="ind">переданный индекс</param>
        /// <returns>Элемент запрашиваемого перечисления</returns>
        private T GetEnum<T>(int ind)
        {
            Type type = typeof(T);
            string[] arr = Enum.GetNames(type);
            return (T)Enum.Parse(type, arr[ind]);
        }
    }

    /// <summary>
    /// Струкрута которая хранит в себе все данные робота необходимые для старта оптимизации / тестирования
    /// </summary>
    struct BotParamKeeper
    {
        /// <summary>
        /// Имя робота
        /// </summary>
        public string BotName;
        /// <summary>
        /// Список входных параметров робота
        /// </summary>
        public SetFileManager BotParams;
        /// <summary>
        /// оптимизационные параметры робота
        /// </summary>
        public OptimisationInputData OptimisationData;
    }
    #endregion
}
