﻿using OptimisationManagerExtention.Model.OptimisationManager;
using System.IO;
using System;
using System.Xml;
using OptimisationManagerExtention.Model.Terminal;
using System.Collections.Generic;

namespace OptimisationManagerExtention.Model.FileReaders
{
    #region OptimisatorSettingsManagerCreatorFabric
    /// <summary>
    /// Абстрактная фабрика
    /// </summary>
    abstract class OptimisatorSettingsManagerCreator
    {
        public abstract OptimisatorSettingsManager Create(string path);
    }

    /// <summary>
    /// Фабрика по созданиюю основного класса по работе с файлами настроек
    /// </summary>
    class MainOptimisatorSettingsManagerCreator : OptimisatorSettingsManagerCreator
    {

        public override OptimisatorSettingsManager Create(string path)
        {
            return new OptimisatorSettingsManager(path);
        }
    }
    #endregion

    /// <summary>
    /// Класс по работе с файлами настроек
    /// </summary>
    class OptimisatorSettingsManager : IDisposable
    {
        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="pathToReportDataDirectory">Путь к папке в которой храним файл с настройками</param>
        /// <param name="pathToTerminalDirectory">Путь к папке ~Terminal</param>
        /// <param name="setFileName">Имя файла с настройками</param>
        public OptimisatorSettingsManager(string pathToReportDataDirectory)
        {
            if (string.IsNullOrEmpty(pathToReportDataDirectory) ||
                string.IsNullOrWhiteSpace(pathToReportDataDirectory))
            {
                throw new ArgumentException("Path to the Directory with report files is incorrect");
            }

            PathToReportDataDirectory = pathToReportDataDirectory;
        }

        #region Fields and document srttings
        /// <summary>
        /// Путь к папке в которой храним файл с настройками
        /// </summary>
        private readonly string PathToReportDataDirectory;
        /// <summary>
        /// Представление файла с настройками оптимизатора ООП
        /// </summary>
        private readonly XmlDocument document = new XmlDocument();
        /// <summary>
        /// Имя файла с настройками
        /// </summary>
        protected readonly string SettingsFileName = "Settings.xml";
        /// <summary>
        /// Формат даты в котором сохраняются настройки
        /// </summary>
        protected readonly string DTFormat = "yyyy.MM.dd";
        #endregion

        #region SetFileParams
        /// <summary>
        /// Список параметров для сохранения
        /// </summary>
        private List<ParamsItem> _setFileParams = new List<ParamsItem>();
        /// <summary>
        /// Получаем и задаем параметры (.set) файла для созранения
        /// </summary>
        public List<ParamsItem> SetFileParams
        {
            get
            {
                if (!document.HasChildNodes)
                    document.Load(Path.Combine(PathToReportDataDirectory, SettingsFileName));
                var data = document["Settings"]["SetFileParams"];

                List<ParamsItem> ans = new List<ParamsItem>();
                foreach (XmlNode item in data.ChildNodes)
                {
                    ans.Add(new ParamsItem(item.Attributes["Name"].Value)
                    {
                        Value = item["Value"].InnerText,
                        Start = item["Start"].InnerText,
                        Step = item["Step"].InnerText,
                        Stop = item["Stop"].InnerText,
                        IsOptimize = Convert.ToBoolean(item["IsOptimize"].InnerText)
                    });
                }

                return ans;
            }
            set { if (value.Count > 0) _setFileParams = value; }
        }
        #endregion

        #region OptimisationInputData
        /// <summary>
        /// Структура OptimisationInputData занесенная для созранения
        /// </summary>
        private OptimisationInputData? _optimisationInputData = null;
        /// <summary>
        /// Получение и сохранение структуры OptimisationInputData
        /// </summary>
        public virtual OptimisationInputData OptimisationInputData
        {
            get
            {
                return new OptimisationInputData
                {
                    Login = StrToUintNullable(GetItem(NodeType.OptimisationInputData, "Login")),
                    ForvardDate = DateTime.ParseExact(GetItem(NodeType.OptimisationInputData, "ForvardDate"), DTFormat, null),
                    IsVisual = Convert.ToBoolean(GetItem(NodeType.OptimisationInputData, "IsVisual")),
                    Deposit = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "Deposit")),
                    Laverage = GetItem(NodeType.OptimisationInputData, "Laverage"),
                    Currency = GetItem(NodeType.OptimisationInputData, "Currency"),
                    DepositIndex = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "DepositIndex")),
                    ExecutionDelayIndex = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "ExecutionDelayIndex")),
                    ModelIndex = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "ModelIndex")),
                    CurrencyIndex = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "CurrencyIndex")),
                    LaverageIndex = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "LaverageIndex")),
                    OptimisationCriteriaIndex = Convert.ToInt32(GetItem(NodeType.OptimisationInputData, "OptimisationCriteriaIndex"))
                };
            }
            set => _optimisationInputData = value;
        }
        #endregion

        #region ConfigCreator_inputData
        /// <summary>
        /// Структура ConfigCreator_inputData занесенная для созранения
        /// </summary>
        private ConfigCreator_inputData? _configInputData = null;
        /// <summary>
        /// Получение и сохранение структуры ConfigCreator_inputData
        /// </summary>
        public virtual ConfigCreator_inputData ConfigCreator_inputData
        {
            get
            {
                return new ConfigCreator_inputData
                {
                    TF = (ENUM_Timeframes)Convert.ToInt32(GetItem(NodeType.ConfigCreator_inputData, "TF")),
                    Login = StrToUintNullable(GetItem(NodeType.ConfigCreator_inputData, "Login")),
                    TerminalID = GetItem(NodeType.ConfigCreator_inputData, "TerminalID"),
                    pathToBot = GetItem(NodeType.ConfigCreator_inputData, "pathToBot"),
                    setFileName = GetItem(NodeType.ConfigCreator_inputData, "setFileName"),
                    Pass = GetItem(NodeType.ConfigCreator_inputData, "Pass"),
                    CertPass = GetItem(NodeType.ConfigCreator_inputData, "CertPass"),
                    Server = GetItem(NodeType.ConfigCreator_inputData, "Server"),
                    Symbol = GetItem(NodeType.ConfigCreator_inputData, "Symbol"),
                    ReportName = GetItem(NodeType.ConfigCreator_inputData, "ReportName"),
                    From = DateTime.ParseExact(GetItem(NodeType.ConfigCreator_inputData, "From"), DTFormat, null),
                    Till = DateTime.ParseExact(GetItem(NodeType.ConfigCreator_inputData, "Till"), DTFormat, null),
                    OptimisationMode = (ENUM_OptimisationMode)Convert.ToInt32(GetItem(NodeType.ConfigCreator_inputData, "OptimisationMode"))
                };
            }
            set => _configInputData = value;
        }
        #endregion

        /// <summary>
        /// Получение элемента из файла сначтройками
        /// </summary>
        /// <param name="NodeName">Тип структуры</param>
        /// <param name="Name">Имя поля</param>
        /// <returns>
        /// Значение поля
        /// </returns>
        public string GetItem(NodeType NodeName, string Name)
        {
            if (!document.HasChildNodes)
                document.Load(Path.Combine(PathToReportDataDirectory, SettingsFileName));

            return document.SelectSingleNode($"/Settings/{NodeName.ToString()}/Item[@Name='{Name}']").InnerText;
        }
        /// <summary>
        /// Преобразование строки в формат uint?
        /// </summary>
        /// <param name="item">данные для перевода</param>
        /// <returns>
        /// двнные в формат uint?
        /// </returns>
        private uint? StrToUintNullable(string item)
        {
            return ((string.IsNullOrEmpty(item) || string.IsNullOrWhiteSpace(item)) ? null : (uint?)Convert.ToUInt32(item));
        }
        /// <summary>
        /// Метод в котором 
        /// 1. Отчищаем ранее загруженный документ с настройками
        /// 2. Сохраняем данные занесенные для сохранения
        /// </summary>
        public virtual void Dispose()
        {
            // Вложенный метод помогающий писать элементы структур
            void WriteItem(XmlTextWriter writer, string Name, string Value)
            {
                writer.WriteStartElement("Item");

                writer.WriteStartAttribute("Name");
                writer.WriteString(Name);
                writer.WriteEndAttribute();

                writer.WriteString(Value);

                writer.WriteEndElement();
            }
            void WriteElement(XmlTextWriter writer, string Node, string Value)
            {
                writer.WriteStartElement(Node);
                writer.WriteString(Value);
                writer.WriteEndElement();
            }

            // сперва отчистить класс зранящий xml разметку файла с настройками
            if (document != null)
                document.RemoveAll();

            // затем проверяем можно ли созранить резуультаты
            if (!_configInputData.HasValue ||
                !_optimisationInputData.HasValue ||
                _setFileParams.Count == 0)
            {
                return;
            }

            using (var xmlWriter = new XmlTextWriter(Path.Combine(PathToReportDataDirectory, SettingsFileName), null))
            {
                xmlWriter.Formatting = Formatting.Indented;
                xmlWriter.IndentChar = '\t';
                xmlWriter.Indentation = 1;

                xmlWriter.WriteStartDocument();

                xmlWriter.WriteStartElement("Settings");

                xmlWriter.WriteStartElement("OptimisationInputData");
                WriteItem(xmlWriter, "Login", _optimisationInputData.Value.Login.ToString());
                WriteItem(xmlWriter, "ForvardDate", _optimisationInputData.Value.ForvardDate.ToString(DTFormat));
                WriteItem(xmlWriter, "IsVisual", _optimisationInputData.Value.IsVisual.ToString());
                WriteItem(xmlWriter, "Deposit", _optimisationInputData.Value.Deposit.ToString());
                WriteItem(xmlWriter, "Laverage", _optimisationInputData.Value.Laverage);
                WriteItem(xmlWriter, "Currency", _optimisationInputData.Value.Currency);
                WriteItem(xmlWriter, "DepositIndex", _optimisationInputData.Value.DepositIndex.ToString());
                WriteItem(xmlWriter, "ExecutionDelayIndex", _optimisationInputData.Value.ExecutionDelayIndex.ToString());
                WriteItem(xmlWriter, "ModelIndex", _optimisationInputData.Value.ModelIndex.ToString());
                WriteItem(xmlWriter, "CurrencyIndex", _optimisationInputData.Value.CurrencyIndex.ToString());
                WriteItem(xmlWriter, "LaverageIndex", _optimisationInputData.Value.LaverageIndex.ToString());
                WriteItem(xmlWriter, "OptimisationCriteriaIndex", _optimisationInputData.Value.OptimisationCriteriaIndex.ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("ConfigCreator_inputData");
                WriteItem(xmlWriter, "TF", ((int)_configInputData.Value.TF).ToString());
                WriteItem(xmlWriter, "Login", _configInputData.Value.Login.ToString());
                WriteItem(xmlWriter, "TerminalID", _configInputData.Value.TerminalID.ToString());
                WriteItem(xmlWriter, "pathToBot", _configInputData.Value.pathToBot);
                WriteItem(xmlWriter, "setFileName", _configInputData.Value.setFileName);
                WriteItem(xmlWriter, "Pass", _configInputData.Value.Pass);
                WriteItem(xmlWriter, "CertPass", _configInputData.Value.CertPass);
                WriteItem(xmlWriter, "Server", _configInputData.Value.Server);
                WriteItem(xmlWriter, "Symbol", _configInputData.Value.Symbol);
                WriteItem(xmlWriter, "ReportName", _configInputData.Value.ReportName);
                WriteItem(xmlWriter, "From", _configInputData.Value.From.ToString(DTFormat));
                WriteItem(xmlWriter, "Till", _configInputData.Value.Till.ToString(DTFormat));
                WriteItem(xmlWriter, "OptimisationMode", ((int)_configInputData.Value.OptimisationMode).ToString());
                xmlWriter.WriteEndElement();

                xmlWriter.WriteStartElement("SetFileParams");
                foreach (var item in _setFileParams)
                {
                    xmlWriter.WriteStartElement("Variable");

                    xmlWriter.WriteStartAttribute("Name");
                    xmlWriter.WriteString(item.Variable);
                    xmlWriter.WriteEndAttribute();

                    WriteElement(xmlWriter, "Value", item.Value);
                    WriteElement(xmlWriter, "Start", item.Start);
                    WriteElement(xmlWriter, "Step", item.Step);
                    WriteElement(xmlWriter, "Stop", item.Stop);
                    WriteElement(xmlWriter, "IsOptimize", item.IsOptimize.ToString());

                    xmlWriter.WriteEndElement();
                }
                xmlWriter.WriteEndElement();

                xmlWriter.WriteEndElement();
                xmlWriter.WriteEndDocument();
                xmlWriter.Close();
            }
        }
    }

    /// <summary>
    /// Тип данных для запрашиваемой структуры 
    /// </summary>
    enum NodeType
    {
        ConfigCreator_inputData,
        OptimisationInputData
    }
}
