﻿using System;
using System.Collections.Generic;
using System.Xml;

namespace OptimisationManagerExtention.Model.FileReaders
{
    #region ReportReaderFabric
    abstract class ReportReaderCreator
    {
        public abstract ReportReader Create(string path);
    }

    class MainReportReaderCreator : ReportReaderCreator
    {
        public override ReportReader Create(string path)
        {
            return new ReportReader(path);
        }
    }
    #endregion

    /// <summary>
    /// Класс читающий отчеты оптимизации записанные в xml дакумент
    /// </summary>
    class ReportReader : IDisposable
    {
        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="path">Путь к документу</param>
        public ReportReader(string path)
        {
            GetDocumentProperties(path);
        }
        /// <summary>
        /// Представление файла с отчетом в формете ООП
        /// </summary>
        private readonly XmlDocument document = new XmlDocument();
        /// <summary>
        /// Еллекция нодов докусента (коллекция строк в excel таблице)
        /// </summary>
        private System.Collections.IEnumerator enumerator;

        #region DocumentProperties and column names
        /// <summary>
        /// Имена колонок документа
        /// </summary>
        protected List<string> columns = new List<string>();
        /// <summary>
        /// Доступ извне к коллекции колонок докусента возвращается копия коллекции 
        /// для того что бы защитить изначальную коллекцию от модификации 
        /// </summary>
        public List<string> ColumnNames => new List<string>(columns);
        /// <summary>
        /// Заголовок документа
        /// </summary>
        public string Title { get; protected set; }
        /// <summary>
        /// Автор документа
        /// </summary>
        public string Author { get; protected set; }
        /// <summary>
        /// Дата создания документа
        /// </summary>
        public DateTime Created { get; protected set; }
        /// <summary>
        /// Сервер на котором производилась оптимизация
        /// </summary>
        public string Server { get; protected set; }
        /// <summary>
        /// Изначальный депозит 
        /// </summary>
        public Deposit InitialDeposit { get; protected set; }
        /// <summary>
        /// Кредитное плече
        /// </summary>
        public int Leverage { get; protected set; }
        #endregion

        /// <summary>
        /// Команда на чтение строки из таблицы с оптимизациями
        /// </summary>
        /// <param name="row">
        /// Прочитанная строке key - заголовок столбца; value - значение ячейки</param>
        /// <returns>
        /// true - если прочтена строка
        /// false - если строка не была прочтена
        /// </returns>
        public virtual bool Read(out List<KeyValuePair<string, object>> row)
        {
            row = new List<KeyValuePair<string, object>>();

            if (enumerator == null)
                return false;

            bool ans = enumerator.MoveNext();
            if (ans)
            {
                XmlNodeList nodes = ((XmlElement)enumerator.Current).ChildNodes;

                for (int i = 0; i < columns.Count; i++)
                {
                    string value = nodes[i]["Data"].InnerText;
                    string type = nodes[i]["Data"].Attributes["ss:Type"].Value;
                    KeyValuePair<string, object> item = new KeyValuePair<string, object>(columns[i], ConvertToType(value, type));
                    row.Add(item);
                }
            }
            return ans;
        }
        /// <summary>
        /// Сброс ридера в изначальное значение
        /// </summary>
        public void ResetReader()
        {
            if (enumerator != null)
            {
                enumerator.Reset(); // Сбрасываем
                enumerator.MoveNext(); // Пропускаем заголовки
            }
        }
        /// <summary>
        /// Перевод полученного значения из строки в указанный в файле формат
        /// </summary>
        /// <param name="value">прочитанное значение</param>
        /// <param name="type">указанный формат в аттребуте Type</param>
        /// <returns>
        /// value приведенное к указанному формату
        /// </returns>
        private object ConvertToType(string value, string type)
        {
            object ans;
            switch (type)
            {
                case "Number":
                    {
                        System.Globalization.NumberFormatInfo provider = new System.Globalization.NumberFormatInfo()
                        {
                            NumberDecimalSeparator = ","
                        };

                        ans = Convert.ToDouble(value.Replace('.', ','), provider);
                    }
                    break;
                case "DateTime": ans = Convert.ToDateTime(value); break;
                case "Boolean":
                    {
                        try
                        {
                            ans = Convert.ToBoolean(value.ToLower());
                        }
                        catch (Exception)
                        {
                            ans = Convert.ToInt32(value) == 1;
                        }
                    }
                    break;
                default: ans = value; break; // String
            }

            return ans;
        }
        /// <summary>
        /// Отчистка поля dicument
        /// </summary>
        public void Dispose()
        {
            document.RemoveAll();
        }
        /// <summary>
        /// Изначальная загрузка документа и установка свойств класса
        /// </summary>
        /// <param name="path">Путь к допусенту</param>
        protected virtual void GetDocumentProperties(string path)
        {
            document.Load(path);

            Title = document["Workbook"]["DocumentProperties"]["Title"].InnerText;
            Author = document["Workbook"]["DocumentProperties"]["Author"].InnerText;
            string DT = document["Workbook"]["DocumentProperties"]["Created"].InnerText;
            Created = Convert.ToDateTime(DT.Replace("Z", ""));
            Server = document["Workbook"]["DocumentProperties"]["Server"].InnerText;
            string[] deposit = document["Workbook"]["DocumentProperties"]["Deposit"].InnerText.Split(' ');
            InitialDeposit = new Deposit(Convert.ToDouble(deposit[0]), deposit[1]);
            Leverage = Convert.ToInt32(document["Workbook"]["DocumentProperties"]["Leverage"].InnerText);

            enumerator = document["Workbook"]["Worksheet"]["Table"].ChildNodes.GetEnumerator();
            enumerator.MoveNext();

            foreach (XmlElement item in (XmlElement)enumerator.Current)
            {
                columns.Add(item["Data"].InnerText);
            }
        }
    }

    /// <summary>
    /// Структура представляющая значение депозита перед началом тестов
    /// </summary>
    struct Deposit
    {
        // Конструктор
        public Deposit(double value, string currency)
        {
            Value = value;
            Currency = currency;
        }
        /// <summary>
        /// Значение 
        /// </summary>
        public double Value { get; }
        /// <summary>
        /// Валюта депозита
        /// </summary>
        public string Currency { get; }
    }
}
