﻿using OptimisationManagerExtention.Model.DirectoryManagers;
using OptimisationManagerExtention.Model.FileReaders;
using OptimisationManagerExtention.Model.Terminal;
using System;
using System.Collections.Generic;
using System.IO;

namespace OptimisationManagerExtention.Model.OptimisationManager.Managers
{
    class SimpleOptimisationManager_Creator : OptimisationManagerFabric
    {
        public SimpleOptimisationManager_Creator(TerminalDirectory terminalDirectory,
                                                 TerminalCreator terminalCreator,
                                                 ConfigCreator configCreator,
                                                 ReportReaderCreator reportReaderCreator,
                                                 SetFileManagerCreator setFileManagerCreator,
                                                 OptimisationExtentionWorkingDirectory currentWorkingDirectory,
                                                 Action<double, string, bool> pbUpdate,
                                                 OptimisatorSettingsManagerCreator optimisatorSettingsManagerCreator) : base("Simple")
        {
            this.terminalDirectory = terminalDirectory;
            this.terminalCreator = terminalCreator;
            this.configCreator = configCreator;
            this.reportReaderCreator = reportReaderCreator;
            this.setFileManagerCreator = setFileManagerCreator;
            this.currentWorkingDirectory = currentWorkingDirectory;
            this.pbUpdate = pbUpdate;
            this.optimisatorSettingsManagerCreator = optimisatorSettingsManagerCreator;
        }

        protected readonly TerminalDirectory terminalDirectory;
        protected readonly TerminalCreator terminalCreator;
        protected readonly ConfigCreator configCreator;
        protected readonly ReportReaderCreator reportReaderCreator;
        protected readonly SetFileManagerCreator setFileManagerCreator;
        protected readonly OptimisationExtentionWorkingDirectory currentWorkingDirectory;
        protected readonly Action<double, string, bool> pbUpdate;
        protected readonly OptimisatorSettingsManagerCreator optimisatorSettingsManagerCreator;

        public override OptimisationManager Create(Dictionary<string, BotParamKeeper> botParamsKeeper,
                                                   List<ViewModel.TerminalAndBotItem> selectedTerminals)
        {
            return new SimpleOptimisationManager(terminalDirectory,
                                                 terminalCreator,
                                                 configCreator,
                                                 reportReaderCreator,
                                                 setFileManagerCreator,
                                                 currentWorkingDirectory,
                                                 botParamsKeeper,
                                                 pbUpdate,
                                                 selectedTerminals,
                                                 optimisatorSettingsManagerCreator);
        }
    }
    class SimpleOptimisationManager : OptimisationManager
    {
        /// <summary>
        /// Конструктор
        /// </summary>
        /// <param name="pathToTerminalFolder">Путь к папки ~/Terminal</param>
        /// <param name="terminalCreator">Фабрика по созданию терминаллов</param>
        /// <param name="configCreator">Фабрика по созданию конфигурационных файлов</param>
        /// <param name="reportReaderCreator">Фабрика по созданию читателей отчетов</param>
        /// <param name="setFileManagerCreator">Фабрика по созданию менеджера файла настроек робота для оптимизации</param>
        /// <param name="currentWorkingDirectory">Класс описывающий струкрурутрабочей папки</param>
        /// <param name="botParamsKeeper">Словарь хранящий в себе параметры робота</param>
        /// <param name="pbUpdate">Метод обновляющий прогресс бар</param>
        /// <param name="selectedTerminals">Массив врбранных терминаллов для оптимизации</param>
        public SimpleOptimisationManager(TerminalDirectory terminalDirectory,
                                         TerminalCreator terminalCreator,
                                         ConfigCreator configCreator,
                                         ReportReaderCreator reportReaderCreator,
                                         SetFileManagerCreator setFileManagerCreator,
                                         OptimisationExtentionWorkingDirectory currentWorkingDirectory,
                                         Dictionary<string, BotParamKeeper> botParamsKeeper,
                                         Action<double, string, bool> pbUpdate,
                                         List<ViewModel.TerminalAndBotItem> selectedTerminals,
                                         OptimisatorSettingsManagerCreator settingsManagerCreator) : base(terminalDirectory,
                                                                                                          terminalCreator,
                                                                                                          configCreator,
                                                                                                          reportReaderCreator,
                                                                                                          setFileManagerCreator,
                                                                                                          currentWorkingDirectory,
                                                                                                          botParamsKeeper,
                                                                                                          pbUpdate,
                                                                                                          selectedTerminals,
                                                                                                          settingsManagerCreator)
        {
        }
        /// <summary>
        /// Деструктор
        /// </summary>
        ~SimpleOptimisationManager() => UnsubscribeTerminals();
        /// <summary>
        /// Массив со списком менеджеров терминаллов запущенных в оптимизационный процесс
        /// </summary>
        private readonly List<ITerminalManager> terminals = new List<ITerminalManager>();
        /// <summary>
        /// Метод прерывающий процесс оптимизации и силой закрывающий терминаллы
        /// </summary>
        public override void BreakOptimisation()
        {
            foreach (var item in terminals)
            {
                if (item.IsActive)
                    item.Close();
            }
        }
        private void UnsubscribeTerminals()
        {
            if (terminals.Count > 0)
            {
                foreach (var item in terminals)
                {
                    item.TerminalClosed -= Terminal_TerminalClosed;
                }
                terminals.Clear();
            }
        }

        protected override void DoOptimisation()
        {
            UnsubscribeTerminals();

            double pb_step = 100.0 / (botParamsKeeper.Count + 1);

            foreach (var item in botParamsKeeper)
            {
                pbUpdate(pb_step, item.Key, false);

                ConfigCreator_inputData configInputData = GetConfigCreator_inputData(item.Key);
                OptimisationInputData optData = item.Value.OptimisationData;

                Config config = CreateConfig(configInputData, optData);

                ITerminalManager terminal = GetTerminal(config, item.Key);
                terminal.TerminalClosed += Terminal_TerminalClosed;
                terminal.Run();

                terminals.Add(terminal);
            }

            pbUpdate(pb_step, "Waiting for Results", false);

            foreach (var item in terminals)
            {
                if (item.IsActive)
                    item.WaitForStop();
            }
        }

        protected virtual void Terminal_TerminalClosed(ITerminalManager terminalManager)
        {
            string FileName = new FileInfo(terminalManager.Config.Tester.Expert).Name.Split('.')[0];

            ConfigCreator_inputData ConfigCreator_inputDat = GetConfigCreator_inputData(terminalManager.TerminalID);
            OptimisationInputData optData = botParamsKeeper[terminalManager.TerminalID].OptimisationData;

            MoveReportToWorkingDirectery(terminalManager, FileName, ConfigCreator_inputDat, optData);
        }

        private ConfigCreator_inputData GetConfigCreator_inputData(string TerminalID)
        {
            ViewModel.TerminalAndBotItem settingsData = selectedTerminals.Find(x => x.TerminalID == TerminalID);
            BotParamKeeper ParamKeeper = botParamsKeeper[TerminalID];

            ConfigCreator_inputData ConfigCreator_inputDat = new ConfigCreator_inputData
            {
                TerminalID = TerminalID,
                pathToBot = ParamKeeper.BotName,
                CertPass = settingsData.CertPass,
                From = settingsData.From,
                Till = settingsData.Till,
                Login = settingsData.Login,
                OptimisationMode = settingsData.GetOptimisationMode,
                Pass = settingsData.Pass,
                Server = settingsData.Server,
                setFileName = botParamsKeeper[TerminalID].BotParams.FileInfo.Name,
                Symbol = settingsData.AssetName,
                TF = settingsData.GetTF,
                ReportName = new FileInfo(ParamKeeper.BotName).Name.Split('.')[0]
            };

            return ConfigCreator_inputDat;
        }
    }
}
