﻿using OptimisationManagerExtention.Model.DirectoryManagers;
using OptimisationManagerExtention.Model.FileReaders;
using OptimisationManagerExtention.Model.Terminal;
using System;
using System.Collections.Generic;
using System.IO;

namespace OptimisationManagerExtention.Model.OptimisationManager
{
    /// <summary>
    /// Интерфейс менеджера оптимизации
    /// </summary>
    abstract class OptimisationManager
    {
        public OptimisationManager(TerminalDirectory terminalDirectory,
                                   TerminalCreator terminalCreator,
                                   ConfigCreator configCreator,
                                   ReportReaderCreator reportReaderCreator,
                                   SetFileManagerCreator setFileManagerCreator,
                                   OptimisationExtentionWorkingDirectory currentWorkingDirectory,
                                   Dictionary<string, BotParamKeeper> botParamsKeeper,
                                   Action<double, string, bool> pbUpdate,
                                   List<ViewModel.TerminalAndBotItem> selectedTerminals,
                                   OptimisatorSettingsManagerCreator optimisatorSettingsManagerCreator)
        {
            this.terminalDirectory = terminalDirectory;
            this.terminalCreator = terminalCreator;
            this.configCreator = configCreator;
            this.reportReaderCreator = reportReaderCreator;
            this.setFileManagerCreator = setFileManagerCreator;
            this.currentWorkingDirectory = currentWorkingDirectory;
            this.pbUpdate = pbUpdate;
            this.optimisatorSettingsManagerCreator = optimisatorSettingsManagerCreator;

            // копируем переданные объекты что бы оборвать их связь с моделью тем самым обезопасив себя от их изменения извне
            this.selectedTerminals = new List<ViewModel.TerminalAndBotItem>(selectedTerminals);
            this.botParamsKeeper = new Dictionary<string, BotParamKeeper>(botParamsKeeper);
        }

        protected readonly TerminalDirectory terminalDirectory;
        protected readonly TerminalCreator terminalCreator;
        protected readonly ConfigCreator configCreator;
        protected readonly ReportReaderCreator reportReaderCreator;
        protected readonly SetFileManagerCreator setFileManagerCreator;
        protected readonly OptimisationExtentionWorkingDirectory currentWorkingDirectory;
        protected readonly Dictionary<string, BotParamKeeper> botParamsKeeper;
        protected readonly Action<double, string, bool> pbUpdate;
        protected readonly List<ViewModel.TerminalAndBotItem> selectedTerminals;
        protected readonly OptimisatorSettingsManagerCreator optimisatorSettingsManagerCreator;

        #region AllOptimisationsFinished event
        /// <summary>
        /// Запуск события окончания процесса оптимизации
        /// </summary>
        private void OnAllOptimisationsFinished()
        {
            AllOptimisationsFinished?.Invoke(this);
        }
        /// <summary>
        /// Событие завершения всех прозодов оптимизации по всем заданным терминаллам и роботам соответственно
        /// </summary>
        public event Action<OptimisationManager> AllOptimisationsFinished;
        #endregion

        /// <summary>
        /// Словарь где:
        /// key - ID терминалла
        /// value - полный путь до робота
        /// </summary>
        public virtual Dictionary<string, string> TerminalAndBotPairs
        {
            get
            {
                Dictionary<string, string> ans = new Dictionary<string, string>();
                foreach (var item in botParamsKeeper)
                {
                    ans.Add(item.Key, item.Value.BotName);
                }
                return ans;
            }
        }
        /// <summary>
        /// Признак того производится ли сейчас оптимизация или же нет.
        /// </summary>
        public bool IsOptimisationOrTestInProcess { get; private set; } = false;
        /// <summary>
        /// Метод запуска теста по событию двойного клика 
        /// </summary>
        /// <param name="TerminalID">ID выбранного терминалла</param>
        /// <param name="pathToBot">Путь к роботу относительно папки экспертов</param>
        /// <param name="row">Строка из таблицы оптимизации</param>
        public virtual void StartTest(ConfigCreator_inputData data,
                                      OptimisationInputData optData)
        {
            pbUpdate(0, "Start Test", true);

            double pb_step = 100.0 / 3;

            IsOptimisationOrTestInProcess = true;

            pbUpdate(pb_step, "Create Config File", false);
            Config config = CreateConfig(data, optData);
            config.Tester.Optimization = ENUM_OptimisationMode.Disabled;
            config.Tester.ShutdownTerminal = false;
            config.DeleteKey(ENUM_SectionType.Tester, "ReplaceReport");
            config.DeleteKey(ENUM_SectionType.Tester, "Report");

            pbUpdate(pb_step, "Create TerminalManager", false);
            ITerminalManager terminal = GetTerminal(config, data.TerminalID);

            pbUpdate(pb_step, "Testing", false);
            terminal.Run();
            terminal.WaitForStop();
            IsOptimisationOrTestInProcess = false;
            pbUpdate(0, null, true);
        }
        /// <summary>
        /// Исспользуется в StartTest для создания конфигурированного менеджера терминаллов
        /// </summary>
        /// <param name="config">Файл конфигурации</param>
        /// <param name="TerminalID">ID терминалла выбранного для оптимизации</param>
        /// <returns>подготовленный менеджер терминалла</returns>
        protected virtual ITerminalManager GetTerminal(Config config, string TerminalID)
        {
            DirectoryInfo TerminalChangebleFolder = terminalDirectory.Terminals.Find(x => x.Name == TerminalID);

            ITerminalManager terminal = terminalCreator.Create(TerminalChangebleFolder);
            terminal.Config = config;

            if (MQL5Connector.MainTerminalID == terminal.TerminalID)
                terminal.Portable = true;

            return terminal;
        }
        /// <summary>
        /// Исспользуется в StartTest
        /// </summary>
        /// <param name="TerminalID">ID терминалла выбранного для оптимизации</param>
        /// <param name="pathToBot">Путь к выбранному роботу относительно папки с экспертами</param>
        /// <param name="setFileName">Имя файла с настройками робота</param>
        /// <returns></returns>
        protected virtual Config CreateConfig(ConfigCreator_inputData data,
                                      OptimisationInputData optData)
        {
            DirectoryInfo termonalChangableFolder = terminalDirectory.Terminals.Find(x => x.Name == data.TerminalID);

            Config config = configCreator.Create(Path.Combine(termonalChangableFolder.GetDirectory("config").FullName, "common.ini"))
                                         .DublicateFile(Path.Combine(currentWorkingDirectory.Configs.FullName, $"{data.TerminalID}.ini"));

            // Заполняем конфигурационный файл
            config.Common.Login = data.Login;
            config.Common.Password = data.Pass;
            config.Common.CertPassword = data.CertPass;
            if (!string.IsNullOrEmpty(data.Server) || !string.IsNullOrWhiteSpace(data.Server))
            {
                try
                {
                    config.Common.Server = new ServerAddressKeeper(data.Server);
                }
                catch (Exception e)
                {
                    System.Windows.MessageBox.Show($"Server adress was incorrect. Your adress is '{data.Server}' but mast have following type 'IPv4:Port'" +
                                                   $"\nError message:\n{e.Message}\n\nStack Trace is {e.StackTrace}");
                    return null;
                }
            }

            bool IsOptimisation = (data.OptimisationMode == ENUM_OptimisationMode.Fast_genetic_based_algorithm ||
                                   data.OptimisationMode == ENUM_OptimisationMode.Slow_complete_algorithm);

            config.Tester.Expert = data.pathToBot;
            config.Tester.ExpertParameters = data.setFileName;
            config.Tester.Symbol = data.Symbol;
            config.Tester.Period = data.TF;
            config.Tester.Login = optData.Login;
            config.Tester.Model = optData.Model;
            config.Tester.ExecutionMode = optData.ExecutionDelay;
            config.Tester.Optimization = data.OptimisationMode;
            config.Tester.FromDate = data.From;
            config.Tester.ToDate = data.Till;
            config.Tester.ForwardMode = ENUM_ForvardMode.Custom;
            config.Tester.ForwardDate = optData.ForvardDate;
            config.Tester.ShutdownTerminal = IsOptimisation;
            config.Tester.Deposit = optData.Deposit;
            config.Tester.Currency = optData.Currency;
            config.Tester.Leverage = optData.Laverage;
            config.Tester.OptimizationCriterion = optData.GetOptimisationCriteria;
            config.Tester.Visual = optData.IsVisual;

            if (IsOptimisation)
            {
                config.Tester.Report = data.ReportName;
                config.Tester.ReplaceReport = true;
            }

            return config;
        }

        /// <summary>
        /// Запуск процесса оптимизации / теста по всем запланированным терминаллам
        /// </summary>
        /// <param name="BotParamsKeeper">Список терминаллов, роботов и параметров по роботам</param>
        /// <param name="PBUpdate">Делегат редактирующий значения полсы загрузки и статуса</param>
        /// <param name="sturtup_status">Ответ из функции - используется в случае если не удалось запустить оптимизацию / тест, 
        /// в данную строку пишется причина неудачного запуска</param>
        /// <returns>true - в случае успеха</returns>
        public void StartOptimisation()
        {
            pbUpdate(0, "Start Optimisation", true);
            IsOptimisationOrTestInProcess = true;

            DoOptimisation();
            OnAllOptimisationsFinished();
            IsOptimisationOrTestInProcess = false;
            pbUpdate(0, null, true);
        }
        protected abstract void DoOptimisation();

        /// <summary>
        /// Метод директивно прерывающий оптимизацют
        /// </summary>
        public abstract void BreakOptimisation();

        protected virtual void MoveReportToWorkingDirectery(ITerminalManager terminalManager,
                                                    string FileName,
                                                    ConfigCreator_inputData ConfigCreator_inputData,
                                                    OptimisationInputData OptimisationInputData)
        {

            FileInfo pathToFile_history = new FileInfo(Path.Combine(terminalManager.TerminalChangeableDirectory.FullName, $"{FileName}.xml"));
            FileInfo pathToFile_forward = new FileInfo(Path.Combine(terminalManager.TerminalChangeableDirectory.FullName, $"{FileName}.forward.xml"));
            int _i = 0;
            while (_i <= 100 && (!pathToFile_history.Exists && !pathToFile_forward.Exists))
            {
                _i++;
                System.Threading.Thread.Sleep(500);
            }

            string botName = new FileInfo(terminalManager.Config.Tester.Expert).Name.Split('.')[0];
            DirectoryInfo terminalReportDirectory = currentWorkingDirectory.Reports.GetDirectory(terminalManager.TerminalID, true);
            if (terminalReportDirectory == null)
                throw new Exception("Can`t create directory");
            DirectoryInfo botReportDir = terminalReportDirectory.GetDirectory(botName, true);
            if (botReportDir == null)
                throw new Exception("Can`t create directory");

            FileInfo _history = new FileInfo(Path.Combine(botReportDir.FullName, "History.xml"));
            FileInfo _forward = new FileInfo(Path.Combine(botReportDir.FullName, "Forward.xml"));

            if (_history.Exists)
                _history.Delete();
            if (_forward.Exists)
                _forward.Delete();

            if (pathToFile_history.Exists)
            {
                pathToFile_history.CopyTo(_history.FullName, true);
                pathToFile_history.Delete();
            }
            if (pathToFile_forward.Exists)
            {
                pathToFile_forward.CopyTo(_forward.FullName, true);
                pathToFile_forward.Delete();
            }

            string pathToSetFile = Path.Combine(terminalManager.TerminalChangeableDirectory
                                                .GetDirectory("MQL5")
                                                .GetDirectory("Profiles")
                                                .GetDirectory("Tester").FullName,
                                                ConfigCreator_inputData.setFileName);

            using (OptimisatorSettingsManager manager =
                   optimisatorSettingsManagerCreator.Create(botReportDir.FullName))
            {
                manager.OptimisationInputData = OptimisationInputData;
                manager.ConfigCreator_inputData = ConfigCreator_inputData;
                manager.SetFileParams = setFileManagerCreator.Create(pathToSetFile, false).Params;
            }
        }
    }

    struct ConfigCreator_inputData
    {
        public ENUM_Timeframes TF;
        public uint? Login;
        public string TerminalID, pathToBot, setFileName,
               Pass, CertPass, Server, Symbol, ReportName;
        public DateTime From, Till;
        public ENUM_OptimisationMode OptimisationMode;
    }
}
