﻿using Microsoft.VisualStudio.TestTools.UnitTesting;
using OptimisationManagerExtention.Model.DirectoryManagers;
using OptimisationManagerExtention.Model.Terminal;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Linq;

namespace OptimisationManagerExtentionTests
{
    [TestClass]
    public class TerminalManagerTest
    {
        private static TerminalDirectory terminalDirectory =
            new TerminalDirectory(Path.Combine(Environment.GetFolderPath(Environment.SpecialFolder.ApplicationData), "MetaQuotes", "Terminal"));

        [ClassInitialize]
        public static void SetUp(TestContext testContext)
        {
            if (terminalDirectory.Terminals.Count == 0)
                Debug.Write("There are no terminals installed on this computer. Ters results is incorrect");
        }

        [TestMethod]
        public void ExpertsTest_Expected_ExpertsNamesFromSelectedTerminal()
        {
            PropertyCollectionsData data = PropertyTest("Experts");
            CollectionAssert.AreEqual(data.expected, data.actual);
        }
        [TestMethod]
        public void IndicatorsTest_Expected_IndicatorsNamesFromSelectedTerminal()
        {
            PropertyCollectionsData data = PropertyTest("Indicators");
            CollectionAssert.AreEqual(data.expected, data.actual);
        }
        
        [TestMethod]
        public void ScriptsTest_Expected_ScriptsNamesFromSelectedTerminal()
        {
            PropertyCollectionsData data = PropertyTest("Scripts");
            CollectionAssert.AreEqual(data.expected, data.actual);
        }
        [TestMethod]
        public void TerminalInstallationDirectoryTest_Expected_TerminalInstallationDirectoryName()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal);

                string expected = File.ReadAllText(selectedTerminal.GetFiles().First(x => x.Name == "origin.txt").FullName);
                string actual = terminal.TerminalInstallationDirectory.FullName;

                Assert.AreEqual(expected, actual);
            }
        }
        [TestMethod]
        public void TerminalChangebleDirectoryTest_Expected_TerminalChangebleDirectoryName()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal);

                string actual = terminal.TerminalChangeableDirectory.FullName;

                Assert.AreEqual(selectedTerminal.FullName, actual);
            }
        }
        [TestMethod]
        public void MQL5DirectoryTest_Expected_MQL5DirectoryName()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal);

                string expected = selectedTerminal.GetDirectory("MQL5").FullName;
                string actual = terminal.MQL5Directory.FullName;

                if (expected == null || actual == null)
                    throw new Exception("Paths are null");

                Assert.AreEqual(expected, actual);
            }
        }
        [TestMethod]
        public void TerminalIDTest_Expected_TerminalID()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal);

                string expected = selectedTerminal.Name;
                string actual = terminal.TerminalID;

                Assert.AreEqual(expected, actual);
            }
        }
        [TestMethod]
        public void RunAndCloseTest_Expected_IfRunIsActiveTrueAndIfCloseIsActiveFalse()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal)
                {
                    Portable = true
                };

                terminal.Run();

                while (!terminal.IsActive)
                {
                    System.Threading.Thread.Sleep(200);
                }

                Assert.IsTrue(terminal.IsActive);
                terminal.WaitForStop();
                Assert.IsFalse(terminal.IsActive);
            }
        }

        [TestMethod]
        public void RunAndCloseTestWithAutomaticClose_Expected_IfRunIsActiveTrueAndIfCloseIsActiveFalse()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal)
                {
                    Portable = true
                };

                terminal.Run();

                while (!terminal.IsActive)
                {
                    System.Threading.Thread.Sleep(200);
                }

                Assert.IsTrue(terminal.IsActive);
                terminal.Close();
                terminal.WaitForStop();
                Assert.IsFalse(terminal.IsActive);
            }
        }

        [TestMethod]
        public void TerminalClosedEventsTest_Expected_TerminalClosedEvents()
        {
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal)
                {
                    Portable = true
                };

                string expected = selectedTerminal.Name;
                string actual = null;

                terminal.TerminalClosed += (ITerminalManager terminalManager) =>
                {
                    actual = terminalManager.TerminalID;
                };

                terminal.Run();
                while (!terminal.IsActive)
                {
                    System.Threading.Thread.Sleep(200);
                }
                terminal.Close();
                terminal.WaitForStop();
                Assert.AreEqual(expected, actual);
            }
        }

        private List<string> GetEX5FilesR(DirectoryInfo path)
        {
            List<string> files = new List<string>();
            IEnumerable<DirectoryInfo> directories = path.GetDirectories();

            files.AddRange(path.GetFiles("*.ex5").Select(x => x.FullName));

            foreach (var item in directories)
                files.AddRange(GetEX5FilesR(item));

            return files;
        }

        struct PropertyCollectionsData
        {
            public List<string> expected;
            public List<string> actual;
        }

        private PropertyCollectionsData PropertyTest(string DirName)
        {
            List<string> Get(TerminalManager terminal)
            {
                List<string> _ans = new List<string>();

                switch (DirName)
                {
                    case "Experts":
                        _ans.AddRange(terminal.Experts.Select(x => Path.Combine(terminal.MQL5Directory.FullName, DirName, x)).ToList());
                        break;
                    case "Indicators":
                        _ans.AddRange(terminal.Indicators.Select(x => Path.Combine(terminal.MQL5Directory.FullName, DirName, x)).ToList());
                        break;
                    case "Scripts":
                        _ans.AddRange(terminal.Scripts.Select(x => Path.Combine(terminal.MQL5Directory.FullName, DirName, x)).ToList());
                        break;
                }

                return _ans;
            };

            PropertyCollectionsData ans = new PropertyCollectionsData();
            if (terminalDirectory.Terminals.Count > 0)
            {
                DirectoryInfo selectedTerminal = GetTerminalDirectory();
                TerminalManager terminal = new TerminalManager(selectedTerminal);
                List<string> expected = GetEX5FilesR(terminal.MQL5Directory.GetDirectory(DirName));

                List<string> actual = Get(terminal);

                ans.expected = expected;
                ans.actual = actual;
            }

            return ans;
        }

        private DirectoryInfo GetTerminalDirectory()
        {
            Random random = new Random();
            int ind = random.Next(0, terminalDirectory.Terminals.Count - 1);

            return terminalDirectory.Terminals[ind];
        }

    }
}
